
// Copyright in this code is held by Dr B. R. Gladman but free direct or
// derivative use is permitted subject to acknowledgement of its origin.  
//
// Dr. B. R. Gladman (gladman@seven77.demon.co.uk).
//
// AES Algorithm Test - Generate local test files for KAT and Monte Carlo
// tests and optionally compare these files with a  reference set of test
// files
//
// usage: aes_gav /a:[12345] /t:[kec] [/c] [/h]  where:
//  /a:[12345] algorithms to be tested
//  /t:[kec]   type of test
//  /c         compare output and reference file(s)
//  /h         use author's byte order (serpent)
// where:
//  1:mars, 2:rc6, 3:rijndael, 4:serpent, 5:twofish
//  k: generate ECB Known Answer Test files  
//  e: generate ECB Monte Carlo Test files
//  c: generate CBC Monte Carlo Test files
//
// The reference files have a ".txt" extension while those generated use 
// ".txt" if in a different directory or ".dat" otherwise. The directory
// paths for files are set below. Note that, when compared with the NIST
// test vector sequences, this code implements one additional (all zero) 
// test vector as the first vector in each set (test 0).

#include <iostream>
#include <iomanip>
#include <fstream>
#include <cctype>
#include <string>

#include "aes_config.h"
#include "aes_defs.h"
#include "aes_aux.h"
#include "mars.h"
#include "rc6.h"
#include "rijndael.h"
#include "serpent.h"
#include "twofish.h"

// Serpent hack to get test agreement because the test vectors for this 
// algorithm assemble the key vectors in the reverse direction in memory

bool    serpent_hack = false;

// Outputs a test vector file header

void header(std::ostream &outf, int type, AES& alg)
{   std::string  str0("=======================================================");
    std::string  str1("Author:    Dr B R Gladman (gladman@seven77.demon.co.uk)");
    std::string  str2("Test:      ");
    std::string  str3(std::string("Algorithm: ") + alg.name());
    std::string  str4("Filename:  ");
    
    str2 += (type < 4 ? "ECB " : "CBC ");
    str4 += (type < 4 ? "ecb_" : "cbc_");

    switch(type)
    {
        case  0:    str2 += "Variable Key Known Answer Tests"; 
                    str4 += "vk.txt"; break;
        case  1:    str2 += "Variable Text Known Answer Tests"; 
                    str4 += "vt.txt"; break;
        case  2:
        case  4:    str2 += "Monte Carlo (Encryption) Tests"; 
                    str4 += "e_m.txt"; break;
        case  3:    
        case  5:    str2 += "Monte Carlo (Decryption) Tests"; 
                    str4 += "d_m.txt"; break;
    }

    outf << str0 << std::endl << str1 << std::endl << str2 << std::endl << str3
         << std::endl << str4 << std::endl << str0;
}

// Test of Electronic Code Book (ECB) Mode with Fixed Key and Variable Text

void ecb_vt(std::ostream &outf, AES& alg)
{   u4byte      i, j;
	u1byte		pt[16], ct[16], key[32], dummy;

	block_clear(key, 32);	// all zero key

    for(i = 0; i < 3; ++i)	// key lengthes 128, 192 and 256
    {
		block_out(0, &dummy, outf, 128 + 64 * i);	// output key length;

		outf << std::endl;

        block_out(3, key, outf, 16 + 8 * i);	// output key value

        alg.set_key(key, 128 + 64 * i);				// set key value

        for(j = 0; j <= 128; ++j)				// 129 test vectors including
        {										// an all zero one 
            block_out(1, &dummy, outf, j);		// output test number 
			block_clear(pt);					// set all zero plain text
            
            if(j)								// set bit (j-1) if j <> 0

                *(pt + (j - 1) / 8) = 0x80 >> (j - 1) % 8;

            block_out(4, pt, outf);				// output plaintext

			if(serpent_hack)					// reverse plaintext
												// block for Serpent
				block_reverse(pt);

            alg.encrypt(pt , ct);					// do encryption

			if(serpent_hack)					// reverse ciphertext
												// block for Serpent
				block_reverse(ct);

            block_out(5, ct, outf);				// output ciphertext
        }
    }
}

// Test of Electronic Code Book (ECB) Mode with Fixed Text and Variable Key

void ecb_vk(std::ostream &outf, AES& alg)
{   u4byte      i, j;
	u1byte		pt[16], ct[16], key[32], dummy;

    block_clear(pt);		// all zero plaintext

    for(i = 0; i < 3; ++i)	// key lengths of 128, 192 and 256 bits
    {
		block_out(0, &dummy, outf, 128 + 64 * i);	// output key length;

		outf << std::endl;

        block_out(4, pt, outf);				// output plaintext	

        for(j = 0; j <= 128 + 64 * i; ++j)	// 129, 193 or 257 tests
        {
            block_out(1, &dummy, outf, j);	// output test number
			block_clear(key, 32);

            if(j)							// set bit (j-1) if j <> 0

                *(key + (j - 1) / 8) = 0x80 >> (j - 1) % 8;

            block_out(3, key, outf, 16 + 8 * i);	// output key value

			if(serpent_hack)				// reverse key block for Serpent
			
				block_reverse(key, 16 + 8 * i);
			
            alg.set_key(key, 128 + 64 * i);		// set key value

            alg.encrypt(pt , ct);				// alg.encrypt

			if(serpent_hack)				// reverse ciphertext for Serpent

				block_reverse(ct);

            block_out(5, ct, outf);			// output ciphertext 
        }
    }
}

// Monte Carlo Encryption Test of Electronic Code Book (ECB) Mode

void ecb_me(std::ostream &outf, AES& alg)
{	u4byte	i, j, k;
	u1byte	pt[16], ct[32], key[32], dummy;

    for(i = 0; i < 3; ++i)	// key lengths of 128, 192 and 256 bits
    {
		block_out(0, &dummy, outf, 128 + 64 * i);	// output key length;

        block_clear(pt); block_clear(key, 32);	// clear initial plaintext 
												// and key blocks
		block_copy(ct + 16, pt);				// put plaintext in upper half
												// of double length buffer
        for(j = 0; j < 400; j++)				// 400 Monte Carlo tests
        {
#if	KEEP_USER_HAPPY

            if(!(j % 40))						// keep user happy
	
				std::cout << std::endl;

			std::cout << '.';
#endif
            block_out(1, &dummy, outf, j);		// output test number
            block_out(3, key, outf, 16 + 8 * i);// output key
            block_out(4, pt, outf);				// output plaintext

			if(serpent_hack)	// reverse both input blocks for Serpent
			{
				block_reverse(key, 16 + 8 * i); block_reverse(ct + 16);
			}

            alg.set_key(key, 128 + 64 * i);		// set the key

            for(k = 0; k < 5000; ++k)		// 10000 encryptions alternating
            {								// upper and lower blocks in ct
                alg.encrypt(ct + 16, ct); alg.encrypt(ct, ct + 16);

            }
			
			// compile next key as defined by NIST

			block_xor(key, ct + 16 - 8 * i, 16 + 8 * i);

			if(serpent_hack)	// put blocks in normal mode for Serpent
			{				
				block_reverse(key, 16 + 8 * i); block_reverse(ct + 16);
			}

            block_out(5, ct + 16, outf);	// output ciphertext
			block_copy(pt, ct + 16);		// copy cipertext as next plaintext
        }
    }	
}

// Monte Carlo Decryption Test of Electronic Code Book (ECB) Mode

void ecb_md(std::ostream &outf, AES& alg)
{	u4byte	i, j, k;
	u1byte	pt[16], ct[32], key[32], dummy;

    for(i = 0; i < 3; ++i)	// key lengths 128, 192 and 256 bits
    {
		block_out(0, &dummy, outf, 128 + 64 * i);	// output key length;

        block_clear(pt);				// clear initial plaintext and key
        block_clear(key, 32);
		block_copy(ct + 16, pt);		// copy plaintext into upper half
										// of double length ciphertext block
        for(j = 0; j < 400; j++)		// 400 Monte Carlo tests
        {
#if	KEEP_USER_HAPPY

            if(!(j % 40))				// keep user happy

				std::cout << std::endl;

			std::cout << '.';
#endif
            block_out(1, &dummy, outf, j);			// output test number
			block_out(3, key, outf, 16 + 8 * i);	// output key
            block_out(5, pt, outf);					// output plaintext

			if(serpent_hack)			// reverse input blocks for Serpent
			{		
				block_reverse(key, 16 + 8 * i); block_reverse(ct + 16);
			}

            alg.set_key(key, 128 + 64 * i);	// set key

            for(k = 0; k < 5000; ++k)	// 10000 decryptions alternating
            {							// upper and lower blocks in ct
                alg.decrypt(ct + 16, ct); alg.decrypt(ct, ct + 16);
            }

			// compile next key as defined by NIST

			block_xor(key, ct + 16 - 8 * i, 16 + 8 * i);

			if(serpent_hack)	// put blocks in normal mode for Serpent
			{
				block_reverse(key, 16 + 8 * i); block_reverse(ct + 16);
			}

            block_out(4, ct + 16, outf);	// output ciphertext
			block_copy(pt, ct + 16);		// set ciphertext as next plaintext
        }
    }
}

// Monte Carlo Encryption Test of Cipher Block Chaining (CBC) Mode

void cbc_me(std::ostream &outf, AES& alg)
{	u4byte	i, j, k;
	u1byte	ct[32], key[32], dummy;

    for(i = 0; i < 3; ++i)	// key lengths 128, 192 and 256 bits
    {
		block_out(0, &dummy, outf, 128 + 64 * i);	// output key length;

        block_clear(key, 32);	// clear key: KEY[0]
        block_clear(ct, 32);	// clear ct:  PT[0], ct + 16: IV[0]

        for(j = 0; j < 400; j++)			// 400 Monte Carlo tests
        {
#if	KEEP_USER_HAPPY
			if(!(j % 40))					// keep user happy

				std::cout << std::endl;

			std::cout << '.';
#endif
            block_out(1, &dummy, outf, j);		// output test number
            block_out(3, key, outf, 16 + 8 * i);// output key
            block_out(2, ct + 16, outf);		// output initialisation vector
            block_out(4, ct, outf);				// output plaintext

			if(serpent_hack)			// reverse key block if Serpent
			{
				block_reverse(key, 16 + 8 * i);
			}

            alg.set_key(key, 128 + 64 * i);	// set key

            for(k = 0; k < 5000; ++k)	// 10000 encryptions, two at a time
            {
                block_xor(ct, ct + 16);	// do CBC chaining

				if(serpent_hack)		// reverse input block for Serpent
				{
					block_reverse(ct);
				}

                alg.encrypt(ct, ct);		// do block encryption

				if(serpent_hack)		// reverse output block for Serpent
				{
					block_reverse(ct);
				}

                block_xor(ct + 16, ct);	// do CBC chaining 

				if(serpent_hack)		// reverse input block for Serpent
				{
					block_reverse(ct + 16);
				}

                alg.encrypt(ct + 16, ct + 16);	// do block encryption

				if(serpent_hack)		// reverse output block for Serpent
				{
					block_reverse(ct + 16);
				}
            }
            
            block_out(5, ct + 16, outf);	// output ciphertext

            if(serpent_hack)				// reverse blocks for Serpent
            {
				block_reverse(ct); block_reverse(ct + 16);
            }

			// compile next key as defined by NIST

			block_xor(key, ct + 16 - 8 * i, 16 + 8 * i);

            if(serpent_hack)	// put blocks in normal order for Serpent
            {
				block_reverse(key, 16 + 8 * i);
				block_reverse(ct); block_reverse(ct + 16);
            }
        }
    }
}

// Monte Carlo Encryption Test of Cipher Block Chaining (CBC) Mode

void cbc_md(std::ostream &outf, AES& alg)
{	u4byte	i, j, k;
	u1byte	pt[16], ct[32], key[32], dummy;

    for(i = 0; i < 3; ++i)	// key lengths 128, 192 and 256 bits
    {
		block_out(0, &dummy, outf, 128 + 64 * i);	// output key length;

		block_clear(key, 32);	// clear key: KEY[0]
		block_clear(ct, 32);	// clear ct: IV[0] ct + 16: CT[0]

        for(j = 0; j < 400; j++)			// 400 Monte Carlo tests
        {
#if	KEEP_USER_HAPPY
			if(!(j % 40))					// keep user happy

				std::cout << std::endl;

			std::cout << '.';
#endif
            block_out(1, &dummy, outf, j);		// output test number
            block_out(3, key, outf, 16 + 8 * i);// output key
            block_out(2, ct, outf);				// output initialisation vector
            block_out(5, ct + 16, outf);		// output ciphertext

			if(serpent_hack)			// reverse key block for Serpent
			{
				block_reverse(key, 16 + 8 * i);
			}

            alg.set_key(key, 128 + 64 * i);	// set key

            for(k = 0; k < 5000; ++k)	// 10000 encryptions, two at a time
            {
				if(serpent_hack)		// reverse input block for Serpent
				{
					block_reverse(ct + 16);
				}

                alg.decrypt(ct + 16, pt);	// do block decryption

				if(serpent_hack)		// reverse input and output blocks
				{						// for Serpent
					block_reverse(ct + 16); block_reverse(pt);
				}

				block_xor(ct, pt);		// do CBC chaining

				if(serpent_hack)		// reverse output block for Serpent
				{
					block_reverse(ct);
				}

                alg.decrypt(ct, pt);		// do block decryption

				if(serpent_hack)		// reverse input and output blocks
				{						// for Serpent
					block_reverse(ct); block_reverse(pt);
				}

				block_xor(ct + 16, pt);	// do CBC chaining
            }
      
            block_out(4, ct + 16, outf);// output plaintext

			if(serpent_hack)	// reverse blocks for Serpent
			{
				block_reverse(ct); block_reverse(ct + 16);
			}

			// compile next key as defined by NIST

			block_xor(key, ct + 16 - 8 * i, 16 + 8 * i);

			if(serpent_hack)	// put blocks in normal order for Serpent
			{
				block_reverse(ct); block_reverse(ct + 16);
				block_reverse(key, 16 + 8 * i);
			}
        }
    }
}

// Synchronise two comparison files if they get out of step

int sync(int nbr, std::ifstream &inf, std::string &str, bool outp)
{   int     ty, nn;

    for(;;)
    {   
        ty = find_line(inf, str);
        
        if(ty < 0)

            return ty;

        if(ty == 1)
        {
            nn = get_dec(str.begin() + 2); 
                
            if(nn >= nbr)
                
                return nn;  
        }
            
        if(outp)
        
            std::cout << std::endl << "  " << str; 
    }
}

// Compare two test vector files

void comp_vecs(std::string &fn1, std::string &fn2)
{   std::ifstream            if1, if2;
    std::string              str1, str2;
    int                 ty1, ty2, no1, no2, err_cnt, np_cnt;
    bool                req;

    err_cnt = np_cnt = 0; req = true;

    if1.open(fn1.c_str(), std::ios::in);	// open first file

    if(!if1.is_open())
    {
        std::cout << std::endl 
			<< "*** 1st file (" << fn1 << ") not found ***"; return;
    }

    if2.open(fn2.c_str(), std::ios::in);

    if(!if2.is_open())
    {
        std::cout << std::endl 
			<< "*** 2nd file (" << fn2 << ") not found ***"; return;
    }

    while(!(if1.eof() && if2.eof()))	// while there is still input
    {
        if(req)	// if another line needs to be input
        {
            ty1 = find_line(if1, str1); ty2 = find_line(if2, str2);
        }

        if(ty1 < 0 && ty2 < 0)		// if end of file on both files

            break;

        if(ty1 < 0 || ty2 < 0)		// if end of file on one file
        {
            std::cout << std::endl << fn1 
				<< (ty1 < 0 ? " short" : " long") << "er than " << fn2; break;
        }

        if(ty1 == 1)	// if 'test number' line in file 1

            no1 = get_dec(str1.begin() + 2); 

        if(ty2 == 1)	// if 'test number' line in file 2

            no2 = get_dec(str2.begin() + 2);

        if(cmp_nocase(str1, str2) == 0)
        {
            req = true; continue;	// if lines are the same continue
        }

        if(ty1 == 1 && ty2 == 1)	// if not the same but both are at a 
        {							// 'test number' line
            np_cnt += abs(no2 - no1); req = false;

            if(no2 < no1)	// extra tests in file 2
            {
                std::cout << std::endl << "extra test(s) in " << fn2 << ':' 
								<< std::endl << "  " << str2;
                no2 = sync(no1, if2, str2, np_cnt < 10); // skip tests in file 2 
            }
        
            if(no1 < no2)	// extra test in file 1
            {
                std::cout << std::endl << "extra test(s) in " << fn1 << ':' 
								<< std::endl << "  " << str1;
                no1 = sync(no2, if1, str1, np_cnt < 10);// skip tests in file 1
            }
        }
        else if(ty1 != ty2)	// cannot synchronise test vector files
        {
            std::cout << std::endl << "*** synchronisation error tests " 
								<< no1 << " and " << no2 << " ***"; 
            return;
        }
        else if(ty1 >= 0)	// test vector mismatch
        {
            std::cout << std::endl << "*** mismatch error test " << no1 << " *** ";
            
            if(++err_cnt < 6)
            
                std::cout << std::endl << "  line 1: " << str1 << std::endl 
								<< "  line 2: " << str2; 
        }
    }

    if(np_cnt && !err_cnt)	// all tests present match

        std::cout << std::endl << "other tests match" << std::endl;

    else if(!err_cnt)		// some failed matches

        std::cout << std::endl << fn1 << " and " << fn2 << " match" << std::endl;
}

// array of functions to call for each test

typedef void (*f_p)(std::ostream&, AES& alg);
f_p f_ptr[6] = { ecb_vk, ecb_vt, ecb_me, ecb_md, cbc_me, cbc_md };

// do the tests for each algorithm

void do_tests(bool do_cmp, bool ttype[3], AES& alg)
{   std::ofstream    outf;
    int         i;
    std::string      name1, name2, extn;

    std::cout << std::endl << "Generate and verify tests for " << alg.name() << std::endl;

    for(i = 0; i < 3; ++i)	// for each type of test /k /e /c (2 tests each)
    {
        if(ttype[i])		// if this test required
        {
            // name of file for output of generated test vectors

			name1 = out_path + std::string(alg.name()) + aes_name[2 * i];

            outf.open(name1.c_str(), std::ios_base::out);

            if(outf)		// if output file is open write it
            {
                header(outf, 2 * i, alg); f_ptr[2 * i](outf, alg); outf << std::endl; 
				outf.close();
        
                if(do_cmp)	// compare it with reference if required
                {
                    name2 = ref_path + std::string(alg.name()) + aes_name[2 * i];
                    comp_vecs(name2, name1);
                }
            }

            // name of file for output of generated test vectors

            name1 = out_path + std::string(alg.name()) + aes_name[2 * i + 1];

            outf.open(name1.c_str(), std::ios_base::out);

            if(outf)		// if output file is open write it
            {
                header(outf, 2 * i + 1, alg); f_ptr[2 * i + 1](outf, alg); 
				outf << std::endl; outf.close();
        
                if(do_cmp)	// compare it with reference if required
                {
                    name2 = ref_path + std::string(alg.name()) + aes_name[2 * i + 1];
                    comp_vecs(name2, name1);
                }
            }
        }
    }
}

void main(int argc, char *argv[])
{	bool        do_cmp, tyf[3];
	mars		_mars;
	rc6			_rc6;
	rijndael	_rijndael;
	serpent		_serpent;
	twofish		_twofish;

    if(argc == 1)
    {
        std::cout << std::endl << "usage: aes_gav /a:[12345] /t:[kec] [/c] [/h]  where:";
        std::cout << std::endl << "  /a:[12345] algorithms to be tested";
        std::cout << std::endl << "  /t:[kec]   type of test";
        std::cout << std::endl << "  /c         compare output and reference file(s)";
        std::cout << std::endl << "  /h         use author's byte order (serpent)";
        std::cout << std::endl << "where:";
        std::cout << std::endl << "  1:mars, 2:rc6, 3:rijndael, 4:serpent, 5:twofish";
        std::cout << std::endl << "  k: generate ECB Known Answer Test files";  
        std::cout << std::endl << "  e: generate ECB Monte Carlo Test files";
        std::cout << std::endl << "  c: generate CBC Monte Carlo Test files";
        std::cout << std::endl << std::endl; exit(0);
    }

    do_cmp = test_args(argc, argv, 'c');
    tyf[0] = test_args(argc, argv, 't', 'k');
    tyf[1] = test_args(argc, argv, 't', 'e');
    tyf[2] = test_args(argc, argv, 't', 'c');

	serpent_hack = false;

	if(test_args(argc, argv, 'a', '1'))

		do_tests(do_cmp, tyf, _mars);

	if(test_args(argc, argv, 'a', '2'))

		do_tests(do_cmp, tyf, _rc6);

	if(test_args(argc, argv, 'a', '3'))

		do_tests(do_cmp, tyf, _rijndael);

	if(test_args(argc, argv, 'a', '4'))
	{
		serpent_hack = test_args(argc, argv, 'h');
			
		do_tests(do_cmp, tyf, _serpent); serpent_hack = false;
	}

	if(test_args(argc, argv, 'a', '5'))

		do_tests(do_cmp, tyf, _twofish);
}
